package com.visibleworkings.trace;

/*
The contents of this file are subject to the Electric Communities E Open 
Source Code License Version 1.0 (the "License"); you may not use this file 
except in compliance with the License. You may obtain a copy of the License 
at http://www.communities.com/EL/.

Software distributed under the License is distributed on an "AS IS" basis, 
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
the specific language governing rights and limitations under the License.

The Original Code is the Distributed E Language Implementation, released 
July 20, 1998. 

The Initial Developer of the Original Code is Electric Communities.  
Copyright (C) 1998 Electric Communities. All Rights Reserved. 

Contributor(s): ______________________________________.
*/
/*
 *  Trace and Logging Package.  Written by Brian Marick,
 *  July-September 1997, for Electric Communities, Inc.
 *  Copyright (c) 1997 by Electric Communities.  All Rights Reserved.
 */

import java.util.Hashtable;
import java.util.Date;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileOutputStream;
import com.visibleworkings.assertion.Assertion;

/**
 * Class that controls storing of TraceMessages in core, for later
 * or concurrent display.
 *
 * The tracebuffer is a linked list of trace messages.  That linked
 * list may be a subset of a larger list.  Just because an old message
 * falls off the end of this list does NOT mean it should be destroyed.
 * Someone else (e.g., the inspector) might be using it.
 */

class TraceBuffer implements TraceMessageAcceptor, TraceConstants {
    /** The first message in the buffer. */
    TraceMessage start;

    /** The last message in the buffer. */
    TraceMessage end;

    /** The max number of messages between start and end (inclusive). */
    int maxSize = STARTING_TRACE_BUFFER_SIZE;

    /** The current number. */
    int currentSize;

    TraceBuffer() {
        // DANGER:  This constructor must be called as part of static
        // initialization of TraceController.  Until that
        // initialization is done, Trace should not be loaded.
        // Therefore, nothing in this constructor should directly or
        // indirectly use a tracing function. 

        // Doing this message initialization this way is pretty icky,
        // but we can't actually *post* this message because 
        // the trace system doesn't fully exist yet.
        TraceMessage message = unpostedTraceMessage("Transient buffer begins.",
                                                    WORLD);

        start = message;
        end = message;
        currentSize = 1;
    }
    public synchronized void accept(TraceMessage message) {
        Assertion.test(start != null);
        Assertion.test(end != null);
        Assertion.test(message != null);

        currentSize++;
        end.next = message;
        end = message;

        // In theory, this loop should only ever execute once.
        while (currentSize > maxSize) {
            start = start.next;
            currentSize--;
        } 

    }
    // Note:  pretty much duplicates TraceLog.changeSize.
    /*package*/ synchronized void changeSize(String value) {
        int newSize;
        if (value.equalsIgnoreCase(DEFAULT_NAME)) {
            newSize = STARTING_TRACE_BUFFER_SIZE; 
        } else if (value.equalsIgnoreCase(UNLIMITED_NAME)) {
            newSize = Integer.MAX_VALUE; 
        } else try { 
            newSize = Integer.parseInt(value);
        } catch (NumberFormatException e) {
            Trace.trace.errorm(
                "Buffer size cannot be changed to illegal value '" +
                value + "'.");
            newSize = maxSize;  // leave unchanged.
        }

        if (newSize < 1) {
            Trace.trace.errorm(
                value + " is too small a threshold size for the log. "
                + "Ignoring.");
            newSize = maxSize;
        }
              
        maxSize = newSize;
        while (currentSize > maxSize) {
            start = start.next;
            currentSize--;
        }
    }

    /**
     * Dump the buffer to a TraceMessageAcceptor.
     */
    /*package*/ synchronized void dump(TraceMessageAcceptor acceptor) {
       Trace.trace.usagem("Dumping internal trace buffer.");

       // We want to demarcate the dumped trace in the new log, but
       // not have the message end up in this log, which could be
       // confusing. 

       TraceMessage traceMessage = unpostedTraceMessage("======================= BEGIN INTERNAL TRACE BUFFER DUMP =======================",
                                                        ERROR);
       acceptor.accept(traceMessage);

       TraceMessage current = start;
       while (current != null) {
           acceptor.accept(current);
           current = current.next;
       }    
       traceMessage = unpostedTraceMessage("======================= END INTERNAL TRACE BUFFER DUMP =======================",
                                           ERROR);
       acceptor.accept(traceMessage);
    }
    // Note:  because messages are never thrown away, 
    // except by GC, it is safe for this not to be synchronized.
    /*package*/ TraceMessage getFirstMessage() {
        Assertion.test(start != null);
        return start;
    }
    // This class does nothing differently before or after setup.
    public void setupIsComplete() {
    }

    /**
     * Create a trace message that isn't posted to all the message
     * acceptors.  Used for messages that have one specific
     * destination.
     */
    private TraceMessage unpostedTraceMessage(String message, int level) {
        return recordTraceMessageXyzzY(message, level);
    }

    /* What's the deal with the funny name? Why does
     * unpostedTraceMessage just call this? It's because we parse the
     * line number for the routine that called a trace method out of
     * the stack. That parsing code (TraceCaller) looks for this
     * method name to start walking the stack. The caller function is
     * expected to be 2 frames above this one's. Hence,
     * "unpostedTraceMessage" fills the same slot as "errorm" -
     * something to be skipped on the way to the true caller.
     */

    private TraceMessage recordTraceMessageXyzzY (String message, int level) {
        TraceMessage tm = new TraceMessage();
        tm.message = message;
        tm.date = new Date();
        tm.object = null;
        tm.level = level;
        tm.subsystem = "trace";
        return tm;
    }
}
