package com.visibleworkings.trace;

/*
The contents of this file are subject to the Electric Communities E Open 
Source Code License Version 1.0 (the "License"); you may not use this file 
except in compliance with the License. You may obtain a copy of the License 
at http://www.communities.com/EL/.

Software distributed under the License is distributed on an "AS IS" basis, 
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
the specific language governing rights and limitations under the License.

The Original Code is the Distributed E Language Implementation, released 
July 20, 1998. 

The Initial Developer of the Original Code is Electric Communities.  
Copyright (C) 1998 Electric Communities. All Rights Reserved. 

Contributor(s): ______________________________________.
*/
/*
 * Original version March 2001 by Brian Marick
 */

import java.io.IOException;
import java.lang.reflect.Field;
import com.visibleworkings.assertion.Assertion;
import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;

/** 
 * This class represents a parsed trace message as read from a log
 * file. It merely breaks the input up into strings.
 *
 * A message's first line looks like the following. The characters are
 * numbered to show where the numbers in the code come from.
 * === 2001/02/22 11:44:11.262 (InstallTest.main:InstallTest.java:15) ERR
 * 0123456789012345678901234567890123456789012345678901234567890123456789
 *           10        20        30        40        50        60
 *
 * Its second line looks like this:
 * InstallTest: Here is a trace message at the ERROR level.
 */
public class TraceParsedMessage {

    public String original; // The original message, suitable for printing.
    public String year;
    public String month;
    public String day;
    public String hour;
    public String minute;
    public String second;
    public String milli;
    public String method;
    public String file;
    public String line;
    public String level;
    public String subsystem;
    public String message;

    private void setField(String fieldName, String value) {
        try { 
            Field field = this.getClass().getDeclaredField(fieldName);
            field.set(this, value);
        } catch (NoSuchFieldException e) {
            Assertion.fail(fieldName + " supposedly does not exist: "
                           + e.getMessage());
        } catch (IllegalAccessException e) {
            Assertion.fail(fieldName + " supposedly does cannot be accessed: "
                           + e.getMessage());
        }
    }

    private void brokennessMessage(String line, String field) throws IOException {
        String problem = "This trace message is garbled around the "
                         + field + " field.";
        problem += System.getProperty("line.separator");
        problem += line;
        throw new IOException(problem);
    }

    private void absoluteField(String line, int start, int pastEnd,
                               String field) throws IOException {
        try {
            setField(field, line.substring(start, pastEnd));
        } catch (IndexOutOfBoundsException e) {
            brokennessMessage(line, field);
        }
    }

    // returns index of ending character.
    private int charEndedField(String line, int start, char endChar,
                               String field) throws IOException {
        try {
            int pastEnd = line.indexOf(endChar, start);
            absoluteField(line, start, pastEnd, field);
            return pastEnd;
        } catch (IndexOutOfBoundsException e) {
            brokennessMessage(line, field);
        }
        return -666; // shuts up compiler
    }

    public TraceParsedMessage(String first, String second) throws IOException {
        original = first + System.getProperty("line.separator") + second;

        absoluteField(first, 4, 8, "year");
        absoluteField(first, 9, 11, "month");
        absoluteField(first, 12, 14, "day");
        absoluteField(first, 15, 17, "hour");
        absoluteField(first, 18, 20, "minute");
        absoluteField(first, 21, 23, "second");
        absoluteField(first, 24, 27, "milli");

        int methodColon = charEndedField(first, 29, ':', "method");
        int fileColon = charEndedField(first, methodColon+1, ':', "file");
        int lineEnd = charEndedField(first, fileColon+1, ')', "line");
        absoluteField(first, lineEnd+2, first.length(), "level");


        int subsystemColon = charEndedField(second, 0, ':', "subsystem");
        absoluteField(second, subsystemColon+2, second.length(), "message");
    }

    /**
     * Answer true if the time represented by this object is within
     * five minutes of the current time. 
     */
    public boolean isRecent() {
        return isRecent(1000 * 60 * 5);
    }


    /**
     * Answer the time represented by this object, in milliseconds.
     */
    public long timeAsMillis() {
        Calendar traceTime =
            new GregorianCalendar(
                Integer.parseInt(year),
                Integer.parseInt(month)-1,
                Integer.parseInt(day),
                Integer.parseInt(hour),
                Integer.parseInt(minute),
                Integer.parseInt(second));
        return traceTime.getTime().getTime();
    }


    /**
     * Answer true if the time represented by this object is within
     * the tolerance, expressed in milliseconds.
     */
    public boolean isRecent(long tolerance) {
        Calendar rightNow = Calendar.getInstance();
        
        long rightNowAsMillis = rightNow.getTime().getTime();
        long traceTimeAsMillis = timeAsMillis();

        // Check if the trace time is in the future.
        if (traceTimeAsMillis > rightNowAsMillis) {
            return false;
        }

        return (rightNowAsMillis - traceTimeAsMillis) < tolerance;
    }

    /**
     * Answer true if I am older than the argument.
     */
    public boolean olderThan(TraceParsedMessage other) {
        return timeAsMillis() < other.timeAsMillis();
    }
}
